@@ -28,7 +28,7 @@ void cutcomment(char*  text)
   //if (cp!=0)
   //  *cp=0;
   while (is_blank(*text))
-    strcpy(text, text +1);
+    memmove(text, text +1, strlen(text +1) +1);
   if (*text == '#')
     *text = 0;
 
@@ -49,7 +49,7 @@ int getsubparam_delim(char*  parameter,
   int j;
   char* cp;
   char* cp2;
-  int size;
+  int len;
 
   cp=(char*)parameter;
   subparam[0]=0;
@@ -61,12 +61,22 @@ int getsubparam_delim(char*  parameter,
   }
   cp2=strchr(cp,delim);
   if (cp2)
-    size=cp2-cp;
+    len=cp2-cp;
   else
-    size=strlen(cp);
-  strncpy(subparam,cp,size);
-  subparam[size]=0;
+    len=strlen(cp);
+
+  // 3.1.7: size_subparam was not used.
+  if (len >= size_subparam)
+    return 0;
+
+  strncpy(subparam,cp,len);
+  subparam[len]=0;
   cutspaces(subparam);
+
+  // 3.1.7:
+  if (!(*subparam))
+    return 0;
+
   return 1;
 }
 
