@@ -18,9 +18,15 @@ Either version 2 of the License, or (at your option) any later version.
 #include <string.h>
 #include <stdarg.h>
 #include <syslog.h>
+#include <ctype.h>
+#ifdef USE_ICONV
+#include <iconv.h>
+#include <errno.h>
+#endif
 #include "charset.h"
 #include "logging.h"
 #include "smsd_cfg.h"
+#include "pdu.h"
 
 // For incoming character 0x24 conversion:
 // Change this if other than Euro character is wanted, like '?' or '$'.
@@ -292,6 +298,11 @@ char iso_8859_15_chars[] =
 	0   , 0
 };
 
+#ifdef USE_ICONV
+static iconv_t iconv4ucs;	// UCS2->UTF8 descriptor
+static iconv_t iconv2ucs;	// UTF8->UCS2 descriptor
+#endif
+
 int special_char2gsm(char ch, char *newch)
 {
   int table_row = 0;
@@ -644,7 +655,7 @@ int iso2utf8_file(FILE *fp, char *ascii, int userdatalength)
         logch("%s ", logtmp);
       }
 
-      if (fwrite(tmp, 1, len, fp) != len)
+      if (fwrite(tmp, 1, len, fp) != (size_t)len)
       {
         if (log_charconv)
           logch(NULL);
@@ -717,6 +728,7 @@ int gsm2iso(char* source, int size, char* destination, int max)
   return dest_count;
 }
 
+#ifndef USE_ICONV
 int decode_ucs2(char *buffer, int len)
 {
   int i;
@@ -748,6 +760,7 @@ int decode_ucs2(char *buffer, int len)
 
   return i;
 }
+#endif
 
 // ******************************************************************************
 // Collect character conversion log, flush it if called with format==NULL.
@@ -807,5 +820,312 @@ char prch(char ch)
   return '.';
 }
 
+//hdr --------------------------------------------------------------------------------
+int iso2utf8(
+	//
+	// Converts to the buffer. Returns -1 in case of error, >= 0 = length of dest.
+	//
+	char *ascii,
+	int userdatalength,
+	size_t ascii_size
+)
+{
+	int result = 0;
+	int idx;
+	unsigned int c;
+	char tmp[10];
+	int len;
+	char logtmp[51];
+	int i;
+	char *buffer;
+
+	if (userdatalength < 0)
+		return -1;
+
+	if (!(buffer = (char *) malloc(ascii_size)))
+		return -1;
+
+#ifdef DEBUGMSG
+	log_charconv = 1;
+#endif
+
+	if (log_charconv)
+	{
+		*logch_buffer = 0;
+		logch("!! iso2utf8(..., userdatalength=%i)", userdatalength);
+		logch(NULL);
+	}
+
+	for (idx = 0; idx < userdatalength; idx++)
+	{
+		len = 0;
+		c = ascii[idx] & 0xFF;
+		// Euro character is 20AC in UTF-8, but A4 in ISO-8859-15:
+		if (c == 0xA4)
+			c = 0x20AC;
+
+		if (c <= 0x7F)
+			tmp[len++] = (char) c;
+		else if (c <= 0x7FF)
+		{
+			tmp[len++] = (char) (0xC0 | ((c >> 6) & 0x1F));
+			tmp[len++] = (char) (0x80 | (c & 0x3F));
+		}
+		else if (c <= 0x7FFF)	// or <= 0xFFFF ?
+		{
+			tmp[len++] = (char) (0xE0 | ((c >> 12) & 0x0F));
+			tmp[len++] = (char) (0x80 | ((c >> 6) & 0x3F));
+			tmp[len++] = (char) (0x80 | (c & 0x3F));
+		}
+
+		if (len == 0)
+		{
+			if (log_charconv)
+				logch(NULL);
+			writelogfile0(LOG_NOTICE, 0, tb_sprintf("UTF-8 conversion error with %i. ch 0x%2X %c.", idx + 1, c, (char) c));
+#ifdef DEBUGMSG
+			printf("%s\n", tb);
+#endif
+		}
+		else
+		{
+			if (log_charconv)
+			{
+				sprintf(logtmp, "%02X[%c]", (unsigned char) ascii[idx], prch(ascii[idx]));
+				if (len > 1 || ascii[idx] != tmp[0])
+				{
+					strcat(logtmp, "->");
+					for (i = 0; i < len; i++)
+						sprintf(strchr(logtmp, 0), "%02X", (unsigned char) tmp[i]);
+				}
+				logch("%s ", logtmp);
+			}
+
+			if ((size_t) (result + len) < ascii_size - 1)
+			{
+				strncpy(buffer + result, tmp, len);
+				result += len;
+			}
+			else
+			{
+				if (log_charconv)
+					logch(NULL);
+				writelogfile0(LOG_NOTICE, 0, tb_sprintf("Fatal error (buffer too small) in UTF-8 conversion"));
+#ifdef DEBUGMSG
+				printf("%s\n", tb);
+#endif
+				result = -1;
+				break;
+			}
+		}
+	}
+
+	if (log_charconv)
+		logch(NULL);
+
+	if (result >= 0)
+	{
+		memcpy(ascii, buffer, result);
+		ascii[result] = 0;
+	}
+
+	free(buffer);
+
+	return result;
+}
+
+//hdr --------------------------------------------------------------------------------
+int encode_7bit_packed(
+	//
+	// Encodes a string to GSM 7bit (USSD) packed format.
+	// Returns number of septets.
+	// Handles padding as defined on GSM 03.38 version 5.6.1 (ETS 300 900) page 17.
+	//
+	char *text,
+	char *dest,
+	size_t size_dest
+)
+{
+	int len;
+	int i;
+	char buffer[512];
+	char buffer2[512];
+	char padding = '\r';
+
+	//len = iso_utf8_2gsm(text, strlen(text), buffer2, sizeof(buffer2), ALPHABET_AUTO, 0);
+	len = iso_utf8_2gsm(text, strlen(text), buffer2, sizeof(buffer2));
+
+#ifdef DEBUGMSG
+	printf("characters: %i\n", len);
+	printf("characters %% 8: %i\n", len % 8);
+#endif
+
+	if ((len % 8 == 7) || (len % 8 == 0 && len && buffer2[len - 1] == padding))
+	{
+		if ((size_t) len < sizeof(buffer2) - 1)
+		{
+			buffer2[len++] = padding;
+#ifdef DEBUGMSG
+			printf("adding padding, characters: %i\n", len);
+#endif
+		}
+	}
+
+	i = text2pdu(buffer2, len, buffer, 0);
+	snprintf(dest, size_dest, "%s", buffer);
+
+#ifdef DEBUGMSG
+	printf("octets: %i\n", strlen(buffer) / 2);
+	for (len = 0; buffer[len]; len += 2)
+		printf("%.2s ", buffer + len);
+	printf("\n");
+#endif
+	return i;
+}
+
+//hdr --------------------------------------------------------------------------------
+int decode_7bit_packed(
+	//
+	// Decodes GSM 7bit (USSD) packed string.
+	// Returns length of dest. -1 in the case or error and "ERROR" in dest.
+	// Handles padding as defined on GSM 03.38 version 5.6.1 (ETS 300 900) page 17.
+	//
+	char *text,
+	char *dest,
+	size_t size_dest
+)
+{
+	int len;
+	int i;
+	char buffer[512];
+	char buffer2[512];
+	char *p;
+	int septets;
+	int padding = '\r';
+
+	snprintf(buffer, sizeof(buffer), "%s", text);
+	while ((p = strchr(buffer, ' ')))
+		memmove(p, p + 1, strlen(p + 1) + 1);
+	for (i = 0; buffer[i]; i++)
+		buffer[i] = toupper((int) buffer[i]);
+
+	i = strlen(buffer);
+	if (i % 2)
+	{
+		snprintf(dest, size_dest, "ERROR");
+		return -1;
+	}
+
+	septets = i / 2 * 8 / 7;
+	snprintf(buffer2, sizeof(buffer2), "%02X%s", septets, buffer);
+
+#ifdef DEBUGMSG
+	printf("septets: %i (0x%02X)\n", septets, septets);
+	printf("septets %% 8: %i\n", septets % 8);
+	printf("%s\n", buffer2);
+#endif
+	memset(buffer, 0, sizeof(buffer));
+	pdu2text(buffer2, buffer, &len, 0, 0, 0, 0, 0);
+
+	if ((septets % 8 == 0 && len && buffer[len - 1] == padding) || (septets % 8 == 1 && len > 1 && buffer[len - 1] == padding && buffer[len - 2] == padding))
+	{
+		len--;
+#ifdef DEBUGMSG
+		printf("removing padding, characters: %i\n", len);
+#endif
+	}
+
+	i = gsm2iso(buffer, len, buffer2, sizeof(buffer2));
+	if (incoming_utf8)
+		i = iso2utf8(buffer2, i, sizeof(buffer2));
+	snprintf(dest, size_dest, "%s", buffer2);
+
+	return i;
+}
+
 // ******************************************************************************
 
+#ifdef USE_ICONV
+int iconv_init(void)
+{
+  // do noy use 'UTF8' alias - it not supported in cygwin/mingw
+  return    (iconv4ucs = iconv_open("UTF-8", "UCS-2BE")) != (iconv_t)-1
+         && (iconv2ucs = iconv_open("UCS-2BE", "UTF-8")) != (iconv_t)-1;
+}
+
+static size_t iconv_convert(iconv_t cd, char *buf, size_t* ilen, size_t maxlen)
+{
+  char tmp[MAXTEXT], *out, *in;
+  size_t olen, rc;
+  const char* err;
+
+  if (!maxlen || !ilen || !*ilen || !buf)
+    return 0;
+
+  // reset conversion descriptor
+  iconv(cd, NULL, NULL, NULL, NULL);
+  err = NULL;
+  in = buf;
+  out = tmp;
+  olen = sizeof(tmp);
+  rc = iconv(cd, &in, ilen, &out, &olen);
+  if (rc == (size_t)-1)
+  {
+    switch (errno)
+    {
+      case E2BIG:
+        err = "Buffer to small";
+        break;
+      case EILSEQ:
+        err = "Invalid sequnce";
+        break;
+      case EINVAL:
+        err = "Incomplete sequence";
+        break;
+      default:
+        err = strerror(errno);
+        break;
+    }
+  }
+  olen = sizeof(tmp) - olen;
+  if (olen >= maxlen)
+  {
+    err = "output buffer too small";
+    olen = maxlen;
+  }
+  memcpy(buf, tmp, olen);
+  if (err != NULL)
+    writelogfile(LOG_NOTICE, 0, "Unicode conversion error: %s", err);
+  return olen;
+}
+
+size_t iconv_utf2ucs(char *buf, size_t len, size_t maxlen)
+{
+  return iconv_convert(iconv2ucs, buf, &len, maxlen);
+}
+
+size_t iconv_ucs2utf(char *buf, size_t len, size_t maxlen)
+{
+  return iconv_convert(iconv4ucs, buf, &len, maxlen);
+}
+
+size_t iconv_ucs2utf_chk(char *buf, size_t len, size_t maxlen)
+{
+  size_t olen, ilen = len;
+  olen = iconv_convert(iconv4ucs, buf, &ilen, maxlen);
+  buf[olen] = 0;
+  if (ilen != 0)
+    ilen = (len - ilen + 1) / 2;
+  return ilen;
+}
+
+int is_ascii_gsm(char* buf, size_t len)
+{
+  char tmp;
+  size_t i;
+  for (i = 0; i < len; i++)
+    if (buf[i] < ' ' || char2gsm(buf[i], &tmp) != 1)
+      return 0;
+  return 1;
+}
+#endif // USE_ICONV
