@@ -385,7 +385,7 @@ int is_highpriority(char *filename)
   char *compare;
   char *compare2 = 0;
 
-  if (ignore_outgoing_priority)
+  if (ignore_outgoing_priority || spool_directory_order)
     return 0;
 
   // get_header() and test_header() can be moved to this file,
@@ -412,7 +412,7 @@ int is_highpriority(char *filename)
       if ((compare2 && strncmp(line, compare2, hlen2) == 0) ||
           strncmp(line, compare, hlen) == 0)
       {
-        cutspaces(strcpy(line, line +hlen));
+        cutspaces(memmove(line, line +hlen, strlen(line +hlen) +1));
         if (!strcasecmp(line,"HIGH"))
           result = 1;
         else if (yesno(line) == 1)
@@ -492,7 +492,7 @@ int getpdufile(char *filename)
   return result;
 }
 
-int getfile(char *dir, char *filename)
+int getfile(int trust_directory, char *dir, char *filename, int lock)
 {
   DIR* dirdata;
   struct dirent* ent;
@@ -530,9 +530,17 @@ int getfile(char *dir, char *filename)
       stat(tmpname, &statbuf);
       if (S_ISDIR(statbuf.st_mode) == 0) /* Is this a directory? */
       {
-        if (strcmp(tmpname +strlen(tmpname) -5, ".LOCK") != 0)
+        // 3.1.7:
+        //if (strcmp(tmpname + strlen(tmpname) - 5, ".LOCK") != 0)
+        i = 1;
+        if (strlen(tmpname) >= 5 && !strcmp(tmpname + strlen(tmpname) - 5, ".LOCK"))
+          i = 0;
+        else if (!strncmp(tmpname, "LOCKED", 6))
+          i = 0;
+
+        if (i)
         {
-          if (trust_outgoing || !islocked(tmpname))
+          if (trust_directory || !islocked(tmpname))
           {
             sprintf(storage_key, "*%s*\n", tmpname);
 
@@ -541,7 +549,7 @@ int getfile(char *dir, char *filename)
               if (!check_access(tmpname))
                 chmod(tmpname, 0766);
 
-            if (!trust_outgoing && !os_cygwin && !file_is_writable(tmpname))
+            if (!trust_directory && !os_cygwin && !file_is_writable(tmpname))
             {
               int result = 1;
               char tmp_filename[PATH_MAX +7];
@@ -580,7 +588,7 @@ int getfile(char *dir, char *filename)
               }
             }
 
-            if (!trust_outgoing && !file_is_writable(tmpname))
+            if (!trust_directory && !file_is_writable(tmpname))
             {
               int report = 1;
               char reason[100];
@@ -626,7 +634,7 @@ int getfile(char *dir, char *filename)
                 int l = strlen(storage_key);
 
                 if ((p = strstr(getfile_err_store, storage_key)))
-                  strncpy(p, p +l, strlen(p) -l +1);
+                  memmove(p, p +l, strlen(p) -l +1);
                 if (!(*getfile_err_store))
                 {
                   free(getfile_err_store);
@@ -667,6 +675,10 @@ int getfile(char *dir, char *filename)
                 strcpy(fname, tmpname);
                 mtime = statbuf.st_mtime;
                 found = 1;
+
+                if (spool_directory_order)
+                  break;
+
               }
 #ifdef DEBUGMSG
               else
@@ -699,11 +711,11 @@ int getfile(char *dir, char *filename)
       p1 = getfile_err_store;
       while ((p2 = strchr(p1, '\n')))
       {
-        strncpy(tmp, p1 +1, p2 -p1 -2);
+        memcpy(tmp, p1 +1, p2 -p1 -2);
         tmp[p2 -p1 -2] = 0;
         //if (access(tmp, F_OK) != 0)
         if (stat(tmp, &statbuf))
-          strncpy(p1, p2 +1, strlen(p2));
+          memmove(p1, p2 +1, strlen(p2));
         else
           p1 = p2 +1;
       }
@@ -720,19 +732,34 @@ int getfile(char *dir, char *filename)
     printf("!! process: %i, getfile_err_store:\n%s", process_id, getfile_err_store);
 #endif
 
-    if (!trust_outgoing && found)
+    // 3.1.9: Lock the file before waiting.
+    if (found && lock)
+      if (!lockfile(fname))
+        found = 0;
+
+    if (!trust_directory && found)
     {
       /* check if the file grows at the moment (another program writes to it) */
       int groesse1;
       int groesse2;
 
-      stat(fname, &statbuf);
-      groesse1=statbuf.st_size;
-      sleep(1);
-      stat(fname, &statbuf);
-      groesse2=statbuf.st_size;
-      if (groesse1 != groesse2)
+      // 3.1.9: check if the file is deleted, or deleted while waiting...
+      if (stat(fname, &statbuf))
         found = 0;
+      else
+      {
+        groesse1 = statbuf.st_size;
+        sleep(1);
+        if (stat(fname, &statbuf))
+          groesse2 = -1;
+        else
+          groesse2 = statbuf.st_size;
+        if (groesse1 != groesse2)
+          found = 0;
+      }
+
+      if (!found && lock)
+        unlockfile(fname);
     }
 
     closedir(dirdata);
@@ -748,100 +775,180 @@ int getfile(char *dir, char *filename)
   return found;
 }
 
-int my_system(char *command, char *info)
+int my_system(
+	//
+	// Executes an external process.
+	//
+	char *command,
+	char *info
+)
 {
-  int pid;
-  int status;
-  time_t start_time;
+	int pid;
+	int status;
+	time_t start_time;
+	char *p;
+	char tmp1[PATH_MAX];
+	char tmp2[PATH_MAX];
 
-  start_time = time(0);
+	// Cannot contain "(" when passed as an argument
+	snprintf(tmp1, sizeof(tmp1), ">%s/smsd_%s_1.XXXXXX", "/tmp", info);
+	while ((p = strchr(tmp1, '(')))
+		*p = '.';
+	while ((p = strchr(tmp1, ')')))
+		*p = '.';
+	while ((p = strchr(tmp1, ' ')))
+		*p = '-';
+
+	snprintf(tmp2, sizeof(tmp2), "2>%s/smsd_%s_2.XXXXXX", "/tmp", info);
+	while ((p = strchr(tmp2, '(')))
+		*p = '.';
+	while ((p = strchr(tmp2, ')')))
+		*p = '.';
+	while ((p = strchr(tmp2, ' ')))
+		*p = '-';
+
+	if (!ignore_exec_output) // 3.1.7
+	{
+		close(mkstemp(tmp1 + 1));
+		close(mkstemp(tmp2 + 2));
+	}
+
+	start_time = time(0);
 #ifdef DEBUGMSG
-  printf("!! my_system(%s, %s)\n", command, info);
+	printf("!! my_system(%s, %s)\n", command, info);
 #endif
-  writelogfile0(LOG_DEBUG, 0, tb_sprintf("Running %s: %s", info, command));
+	writelogfile0(LOG_DEBUG, 0, tb_sprintf("Running %s: %s", info, command));
 
-  pid = fork();
-  if (pid == -1)
-  {
-    writelogfile0(LOG_CRIT, 0, tb_sprintf("Fatal error: fork failed."));
-    return -1;
-  }
+	pid = fork();
+	if (pid == -1)
+	{
+		writelogfile0(LOG_CRIT, 0, tb_sprintf("Fatal error: fork failed."));
+		return -1;
+	}
 
-  if (pid == 0) // only executed in the child
-  {
-    char *argv[4];
-    char *p;
+	if (pid == 0)		// only executed in the child
+	{
+		char *argv[4];
+		char *cmd = 0;
 
 #ifdef DEBUGMSG
-  printf("!! pid=%i, child running external command\n", pid);
+		printf("!! pid=%i, child running external command\n", pid);
 #endif
 
-    // TODO: sh still ok?
-    argv[0] = "sh";
-    if ((p = strrchr(shell, '/')))
-      argv[0] = p +1;
+		// TODO: sh still ok?
+		argv[0] = "sh";
+		if ((p = strrchr(shell, '/')))
+			argv[0] = p + 1;
 
-    argv[1] = "-c";
-    argv[2] = command; //(char*) command;
-    argv[3] = 0;
+		argv[1] = "-c";
+		argv[2] = command;	//(char*) command;
 
-    // 3.1.5:
-    //execv("/bin/sh",argv);  // replace child with the external command
-    execv(shell, argv); // replace child with the external command
-    writelogfile0(LOG_CRIT, 1, tb_sprintf("Fatal error: execv( %s ) returned: %i, %s", shell, errno, strerror(errno)));
+		if (!ignore_exec_output)
+		{
+			if ((cmd = (char *) malloc(strlen(command) + strlen(tmp1) + strlen(tmp2) + 3)))
+			{
+				sprintf(cmd, "%s %s %s", command, tmp1, tmp2);
+				argv[2] = cmd;
+//writelogfile0(LOG_CRIT, 0, tb_sprintf("Running: %s", cmd));
+			}
+		}
+
+		argv[3] = 0;
+
+		// 3.1.5:
+		//execv("/bin/sh",argv);  // replace child with the external command
+		execv(shell, argv);	// replace child with the external command
+		writelogfile0(LOG_CRIT, 1, tb_sprintf("Fatal error: execv( %s ) returned: %i, %s", shell, errno, strerror(errno)));
+
+		free(cmd);
 
 #ifdef DEBUGMSG
-  printf("!! pid=%i, execv() failed, %i, %s\n", pid, errno, strerror(errno));
-  printf("!! child exits now\n");
+		printf("!! pid=%i, execv() failed, %i, %s\n", pid, errno, strerror(errno));
+		printf("!! child exits now\n");
 #endif
-    exit((errno)? errno : 1); // exit with error when the execv call failed
-  }
+		exit((errno) ? errno : 1);	// exit with error when the execv call failed
+	}
 
-  errno = 0;
+	errno = 0;
 #ifdef DEBUGMSG
-  printf("!! father waiting for child %i\n", pid);
+	printf("!! father waiting for child %i\n", pid);
 #endif
-  strcpy(run_info, info);
+	snprintf(run_info, sizeof(run_info), "%s", info);
 
-  while (1)
-  {
-    if (waitpid(pid, &status, 0) == -1)
-    {
-      if (errno != EINTR)
-      {
-        *run_info = 0;
-        writelogfile0(LOG_ERR, 0, tb_sprintf("Done: %s, execution time %i sec., errno: %i, %s", info, time(0) -start_time, errno, strerror(errno)));
-        return -1;
-      }
-    }
-    else
-    {
-      int level = LOG_DEBUG;
-      int trouble = 0;
+	while (1)
+	{
+		if (waitpid(pid, &status, 0) == -1)
+		{
+			if (errno != EINTR)
+			{
+				*run_info = 0;
+				writelogfile0(LOG_ERR, 0, tb_sprintf("Done: %s, execution time %i sec., errno: %i, %s", info, time(0) - start_time, errno, strerror(errno)));
+				return -1;
+			}
+		}
+		else
+		{
+			int level = LOG_DEBUG;
+			int trouble = 0;
 
-      *run_info = 0;
+			*run_info = 0;
 
-      // 3.1.6: When running checkhandler and it spooled a message, return value 2 SHOULD NOT activate trouble logging:
-      //writelogfile0((status == 0) ? LOG_DEBUG : LOG_ERR, (status == 0) ? 0 : 1, tb_sprintf("Done: %s, execution time %i sec., status: %i", info, time(0) - start_time, status));
-      if (!strcmp(info, "checkhandler"))
-      {
-        if (status != 0 && WEXITSTATUS(status) != 2)
-        {
-          level = LOG_ERR;
-          trouble = 1;
-        }
-      }
-      else if (status != 0)
-      {
-        level = LOG_ERR;
-        trouble = 1;
-      }
+			// 3.1.6: When running checkhandler and it spooled a message, return value 2 SHOULD NOT activate trouble logging:
+			//writelogfile0((status == 0) ? LOG_DEBUG : LOG_ERR, (status == 0) ? 0 : 1, tb_sprintf("Done: %s, execution time %i sec., status: %i", info, time(0) - start_time, status));
+			if (!strcmp(info, "checkhandler"))
+			{
+				if (status != 0 && WEXITSTATUS(status) != 2)
+				{
+					level = LOG_ERR;
+					trouble = 1;
+				}
+			}
+			else if (status != 0)
+			{
+				level = LOG_ERR;
+				trouble = 1;
+			}
 
-      writelogfile0(level, trouble, tb_sprintf("Done: %s, execution time %i sec., status: %i (%i)", info, time(0) - start_time, status, WEXITSTATUS(status)));
+			writelogfile0(level, trouble, tb_sprintf("Done: %s, execution time %i sec., status: %i (%i)", info, time(0) - start_time, status, WEXITSTATUS(status)));
 
-      return WEXITSTATUS(status);
-    }
-  }
+			if (!ignore_exec_output)
+			{
+				struct stat statbuf;
+				FILE *fp;
+				char line[2048];
+				int i;
+				char *p;
+
+				for (i = 1; i <= 2; i++)
+				{
+					p = (i == 1) ? tmp1 + 1 : tmp2 + 2;
+
+					if (stat(p, &statbuf) == 0)
+					{
+						if (statbuf.st_size)
+						{
+							writelogfile0(LOG_ERR, 1, tb_sprintf("Exec: %s %s:", info, (i == 1) ? "said something" : "encountered errors"));
+							if ((fp = fopen(p, "r")))
+							{
+								while (fgets(line, sizeof(line), fp))
+								{
+									while (strlen(line) > 1 && strchr("\r\n", line[strlen(line) - 1]))
+										line[strlen(line) - 1] = 0;
+
+									writelogfile0(LOG_ERR, 1, tb_sprintf("! %s", line));
+								}
+								fclose(fp);
+							}
+						}
+
+						unlink(p);
+					}
+				}
+			}
+
+			return WEXITSTATUS(status);
+		}
+	}
 }
 
 int write_pid( char* filename)
@@ -905,8 +1012,7 @@ int parse_validity(char *value, int defaultvalue)
     // Convert to lowercase so upcase is also accepted.
 
     *buffer = 0;
-    strncpy(tmp, value, sizeof(tmp) -1);
-    tmp[sizeof(tmp) -1] = 0;
+    snprintf(tmp, sizeof(tmp), "%s", value);
     cutspaces(tmp);
     for (idx = 0; tmp[idx]; idx++)
     {
@@ -1223,3 +1329,50 @@ unsigned long long time_usec()
 
   return (unsigned long long)tv.tv_sec *1000000 +tv.tv_usec;
 }
+
+int make_datetime_string(char *dest, size_t dest_size, char *a_date, char *a_time, char *a_format)
+{
+  int result = 0;
+  time_t rawtime;
+  struct tm *timeinfo;
+
+  time(&rawtime);
+
+  if (!a_date && !a_time)
+    return strftime(dest, dest_size, (a_format)? a_format : datetime_format, localtime(&rawtime));
+
+  if (a_date && strlen(a_date) >= 8 && a_time && strlen(a_time) >= 8)
+  {
+    timeinfo = localtime(&rawtime);
+    timeinfo->tm_year = atoi(a_date) + 100;
+    timeinfo->tm_mon = atoi(a_date + 3) - 1;
+    timeinfo->tm_mday = atoi(a_date + 6);
+    timeinfo->tm_hour = atoi(a_time);
+    timeinfo->tm_min = atoi(a_time + 3);
+    timeinfo->tm_sec = atoi(a_time + 6);
+    // ?? mktime(timeinfo);
+    result = strftime(dest, dest_size, (a_format)? a_format : datetime_format, timeinfo);
+  }
+
+  return result;
+}
+
+void strcat_realloc(char **buffer, char *str, char *delimiter)
+{
+  int delimiter_length = 0;
+
+  if (delimiter)
+    delimiter_length = strlen(delimiter);
+
+  if (*buffer == 0)
+  {
+    if ((*buffer = (char *) malloc(strlen(str) + delimiter_length + 1)))
+      **buffer = 0;
+  }
+  else
+    *buffer = (char *) realloc((void *) *buffer, strlen(*buffer) + strlen(str) + delimiter_length + 1);
+
+  if (*buffer)
+    sprintf(strchr(*buffer, 0), "%s%s", str, (delimiter) ? delimiter : "");
+}
+
