@@ -35,6 +35,14 @@ Either version 2 of the License, or (at your option) any later version.
 #include <syslog.h>
 #include <sys/ioctl.h>
 #include <errno.h>
+
+#ifndef DISABLE_INET_SOCKET
+// HG
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <netdb.h>
+#endif
+
 #include "extras.h"
 #include "modeminit.h"
 #include "smsd_cfg.h"
@@ -327,67 +335,113 @@ int write_to_modem(char *command, int timeout, int log_command, int print_error)
   int x=0;
   struct termios tio;
 
-  tcgetattr(modem_handle, &tio);
-
   if (command && command[0])
   {
-    if (tio.c_cflag & CRTSCTS)
-    {
-      ioctl(modem_handle, TIOCMGET, &status);
-      while (!(status & TIOCM_CTS))
-      {
-        usleep(100000);
-        timeoutcounter++;
-        ioctl(modem_handle, TIOCMGET, &status);
-        if (timeoutcounter>timeout)
-        {
-          if (print_error)
-            printf("\nModem is not clear to send.\n");
-          else
-          {
-            writelogfile0(LOG_ERR, 1, tb_sprintf("Modem is not clear to send"));
-            alarm_handler0(LOG_ERR, tb);
-          }
-          return 0;
-        }
-      }
-    }
     if (log_command)
       writelogfile(LOG_DEBUG, 0, "-> %s",command);
 
-    // 3.1.5:
-    if (DEVICE.send_delay < 1)
+    // 3.1.9:
+    if (DEVICE.send_handshake_select)
     {
-      if (write(modem_handle, command, strlen(command)) != strlen(command))
+      size_t r = 0, bs, n;
+      ssize_t got;
+      fd_set writefds;
+
+      n = strlen(command);
+      while (n > 0)
       {
-        writelogfile0(LOG_ERR, 1, tb_sprintf("Could not send string, cause: %s", strerror(errno)));
-        alarm_handler0(LOG_ERR, tb);
-        return 0;
-      }
+        bs = (DEVICE.send_delay < 1) ? n : 1;
+        got = write(modem_handle, command + r, bs);
+        if (got < 0)
+        {
+          if (errno == EINTR)
+            continue;
+
+          if (errno != EAGAIN)
+          {
+            writelogfile0(LOG_ERR, 1, tb_sprintf("write_to_modem: error %d: %s", errno, strerror(errno)));
+            alarm_handler0(LOG_ERR, tb);
+            return 0;
+          }
+
+          writelogfile0(LOG_DEBUG, 0, tb_sprintf("write_to_modem: device busy, waiting"));
+          alarm_handler0(LOG_DEBUG, tb);
+          FD_ZERO(&writefds);
+          FD_SET(modem_handle, &writefds);
+          select(modem_handle + 1, NULL, &writefds, NULL, NULL);
+          continue;
+        }
+
+        n -= got;
+        r += got;
+
+        if (DEVICE.send_delay > 0)
+          usleep(DEVICE.send_delay * 1000);
 
-      if (DEVICE.send_delay < 0)
         tcdrain(modem_handle);
+      }
     }
     else
     {
-      for(x=0;x<strlen(command);x++)
+      tcgetattr(modem_handle, &tio);
+
+      if (!DEVICE_IS_SOCKET && tio.c_cflag & CRTSCTS)
       {
-        if (write(modem_handle, command +x, 1) < 1)
+        ioctl(modem_handle, TIOCMGET, &status);
+        while (!(status & TIOCM_CTS))
         {
-          if (print_error)
-            printf("\nCould not send character %c, cause: %s\n",command[x],strerror(errno));
-          else
+          usleep(100000);
+          timeoutcounter++;
+          ioctl(modem_handle, TIOCMGET, &status);
+          if (timeoutcounter>timeout)
           {
-            writelogfile0(LOG_ERR, 1, tb_sprintf("Could not send character %c, cause: %s", command[x], strerror(errno)));
-            alarm_handler0(LOG_ERR, tb);
+            if (print_error)
+              printf("\nModem is not clear to send.\n");
+            else
+            {
+              writelogfile0(LOG_ERR, 1, tb_sprintf("Modem is not clear to send"));
+              alarm_handler0(LOG_ERR, tb);
+            }
+            return 0;
           }
+        }
+      }
+
+      // 3.1.5:
+      if (DEVICE.send_delay < 1)
+      {
+        if ((size_t)write(modem_handle, command, strlen(command)) != strlen(command))
+        {
+          writelogfile0(LOG_ERR, 1, tb_sprintf("Could not send string, cause: %s", strerror(errno)));
+          alarm_handler0(LOG_ERR, tb);
           return 0;
         }
-        usleep(DEVICE.send_delay *1000);
-        tcdrain(modem_handle);
+
+        if (DEVICE.send_delay < 0)
+          tcdrain(modem_handle);
+      }
+      else
+      {
+        for(x=0;(size_t)x<strlen(command);x++)
+        {
+          if (write(modem_handle, command +x, 1) < 1)
+          {
+            if (print_error)
+              printf("\nCould not send character %c, cause: %s\n",command[x],strerror(errno));
+            else
+              {
+              writelogfile0(LOG_ERR, 1, tb_sprintf("Could not send character %c, cause: %s", command[x], strerror(errno)));
+              alarm_handler0(LOG_ERR, tb);
+            }
+            return 0;
+          }
+          usleep(DEVICE.send_delay *1000);
+          tcdrain(modem_handle);
+        }
       }
     }
   }
+
   return 1;
 }
 
@@ -425,6 +479,30 @@ int read_from_modem(char *answer, int max, int timeout)
     }
     else  
     {
+      // 3.1.9
+#ifdef DEBUGMSG
+      log_read_from_modem = 1;
+#endif
+      if (log_read_from_modem)
+      {
+        char tmp[SIZE_LOG_LINE];
+        int i;
+
+        snprintf(tmp, sizeof(tmp), "read_from_modem: count=%i, got=%i:", count, got);
+        for (i = count; i < count + got; i++)
+        {
+          if (strlen(tmp) >= sizeof(tmp) - 6)
+          {
+            strcpy(tmp, "ERROR: too much data");
+            break;
+          }
+
+          sprintf(strchr(tmp, 0), " %02X[%c]", (unsigned char) answer[i], ((unsigned char) answer[i] >= ' ') ? answer[i] : '.');
+        }
+
+        writelogfile(LOG_CRIT, 0, tmp);
+      }
+
       // restart timout counter
       timeoutcounter=0;
       // append a string termination character
@@ -477,8 +555,11 @@ int detect_routed_message(char *answer)
 
     if (p1)
     {
-      writelogfile(LOG_ERR, DEVICE.unexpected_input_is_trouble, "Routed %s detected:\n%s", (is_sr)? "status report" : "message", p1);
-      send_ack = 1;
+      if (!is_sr || !DEVICE.using_routed_status_report)
+        writelogfile(LOG_ERR, DEVICE.unexpected_input_is_trouble, "Routed %s detected:\n%s", (is_sr) ? "status report" : "message", p1);
+
+      if (DEVICE.routed_status_report_cnma)
+        send_ack = 1;
 
       while (p1)
       {
@@ -523,8 +604,13 @@ int detect_routed_message(char *answer)
         p = (is_sr)? "status report" : "message";
         if (can_handle)
         {
-          writelogfile0(LOG_ERR, 0, tb_sprintf("Saved routed %s for later handling. However, you MUST DISABLE %s routing with modem settings.", p, p));
-          alarm_handler0(LOG_ERR, tb);
+          if (is_sr && DEVICE.using_routed_status_report)
+            writelogfile0(LOG_INFO, 0, tb_sprintf("Saved routed %s for later handling.", p));
+          else
+          {
+            writelogfile0(LOG_ERR, 0, tb_sprintf("Saved routed %s for later handling. However, you MUST DISABLE %s routing with modem settings.", p, p));
+            alarm_handler0(LOG_ERR, tb);
+          }
         }
         else
         {
@@ -546,8 +632,7 @@ int detect_routed_message(char *answer)
       char loganswer[1024];
 
       writelogfile(LOG_DEBUG, 0, "Sending acknowledgement");
-      if (write_to_modem("AT+CNMA\r", 30, 1, 0) == 0)
-        ;
+      write_to_modem("AT+CNMA\r", 30, 1, 0);
 
       *loganswer = 0;
       read_from_modem(loganswer, sizeof(loganswer), 2);
@@ -561,11 +646,55 @@ int detect_routed_message(char *answer)
   return result;
 }
 
+void handlephonecall_clip(char *answer)
+{
+	char *p, *e_start, *e_end;
+	int len;
+	char entry_number[SIZE_PB_ENTRY];
+	int entry_type;
+
+	*entry_number = 0;
+	entry_type = 0;
+
+	if ((p = strstr(answer, "+CLIP:")))
+	{
+		if ((e_start = strchr(p, '"')))
+		{
+			e_start++;
+			if ((e_end = strchr(e_start, '"')))
+			{
+		                if ((len = e_end -e_start) < SIZE_PB_ENTRY)
+				{
+					sprintf(entry_number, "%.*s", len, e_start);
+					cutspaces(entry_number);
+					if (*entry_number == '+')
+						memmove(entry_number, entry_number +1, strlen(entry_number +1) +1);
+
+					if (strlen(e_end) >= 3)
+					{
+						e_end += 2;
+						writelogfile(LOG_INFO, 0, "Got phonecall from %s", entry_number);
+						savephonecall(entry_number, atoi(e_end), "");
+					}
+				}
+			}
+		}
+	}
+}
+
 // 3.1beta7: Not waiting any answer if answer is NULL. Return value is then 1/0.
 // 3.1.5: In case of timeout return value is -2.
+
 int put_command(char *command, char *answer, int max, int timeout_count, char *expect)
+{
+
+  return put_command0(command, answer, max, timeout_count, expect, 0);
+}
+
+int put_command0(char *command, char *answer, int max, int timeout_count, char *expect, int silent)
 {
   char loganswer[SIZE_LOG_LINE];
+  char tmpanswer[SIZE_LOG_LINE];
   int timeoutcounter = 0;
   regex_t re;
   int got_timeout = 0;
@@ -605,27 +734,33 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
 
     if (!i && *loganswer && DEVICE.detect_unexpected_input)
     {
-      cutspaces(loganswer);
-      cut_emptylines(loganswer);
+      if (!log_unmodified)
+      {
+        cutspaces(loganswer);
+        cut_emptylines(loganswer);
 
-      if (log_single_lines)
-        change_crlf(loganswer, ' ');
+        if (log_single_lines)
+          change_crlf(loganswer, ' ');
+      }
 
       if (*loganswer)
       {
         // Some modems send unsolicited result code even when status report is stored for future
         // reading. This and SMS-DELIVER indication is not logged.
         if (!strstr(loganswer, "+CDSI:") && !strstr(loganswer, "+CMTI:"))
-          writelogfile(LOG_ERR, DEVICE.unexpected_input_is_trouble, "Unexpected input: %s", loganswer);
+          if (!(strstr(loganswer, "+CLIP:") && DEVICE.phonecalls == 2))
+            writelogfile(LOG_ERR, DEVICE.unexpected_input_is_trouble, "Unexpected input: %s", loganswer);
 
         if (strstr(loganswer, "RING"))
         {
           if (DEVICE.hangup_incoming_call == 1 ||
-              (DEVICE.hangup_incoming_call == -1 && hangup_incoming_call == 1))
+              (DEVICE.hangup_incoming_call == -1 && hangup_incoming_call == 1) ||
+              DEVICE.phonecalls == 2)
           {
             char *command = "AT+CHUP\r";
 
-            if (DEVICE.voicecall_hangup_ath)
+           if (DEVICE.voicecall_hangup_ath == 1 ||
+                (DEVICE.voicecall_hangup_ath == -1 && voicecall_hangup_ath == 1))
               command = "ATH\r";
 
             writelogfile(LOG_NOTICE, 0, "Ending incoming call: %s", loganswer);
@@ -633,27 +768,30 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
             write_to_modem(command, 30, 1, 0);
 
             timeoutcounter = 0;
-            *loganswer = 0;
+            *tmpanswer = 0;
 
             do
             {
-              read_from_modem(loganswer, sizeof(loganswer), 2);  // One read attempt is 200ms
+              read_from_modem(tmpanswer, sizeof(tmpanswer), 2);  // One read attempt is 200ms
 
               // Any answer is ok:
-              if (*loganswer)
+              if (*tmpanswer)
                 break;
 
               timeoutcounter++;;
             }
             while (timeoutcounter < 5);
 
-            cutspaces(loganswer);
-            cut_emptylines(loganswer);
+            if (!log_unmodified)
+            {
+              cutspaces(tmpanswer);
+              cut_emptylines(tmpanswer);
 
-            if (log_single_lines)
-              change_crlf(loganswer, ' ');
+              if (log_single_lines)
+                change_crlf(tmpanswer, ' ');
+            }
 
-            writelogfile(LOG_DEBUG, 0, "<- %s", loganswer);
+            writelogfile(LOG_DEBUG, 0, "<- %s", tmpanswer);
 
             last_command_ended = time_usec();
 
@@ -662,6 +800,10 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
                 usleep_until(last_command_ended +DEVICE.communication_delay *1000);
           }
         }
+
+	if (strstr(loganswer, "+CLIP:") && DEVICE.phonecalls == 2)
+		handlephonecall_clip(loganswer);
+
       }
     }
   }
@@ -685,10 +827,14 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
   }
  
   if (!answer)
-    writelogfile(LOG_DEBUG, 0, "Command is sent");
+  {
+    if (!silent)
+      writelogfile(LOG_DEBUG, 0, "Command is sent");
+  }
   else
   {
-    writelogfile(LOG_DEBUG, 0, "Command is sent, waiting for the answer");
+    if (!silent)
+      writelogfile(LOG_DEBUG, 0, "Command is sent, waiting for the answer");
 
     // wait for the modem-answer 
     answer[0] = 0;
@@ -743,7 +889,7 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
 
     // 3.1.5: Some modems (like Westermo GDW-11) start answer with extra <CR><LF>, check and remove it:
     while (!got_timeout && !strncmp(answer, "\r\n", 2))
-      strcpy(answer, answer +2);
+      memmove(answer, answer +2, strlen(answer +2) +1);
 
     // 3.1.4: error explanation should be included in the answer.
     if (!got_timeout && strstr(answer, "ERROR"))
@@ -756,14 +902,16 @@ int put_command(char *command, char *answer, int max, int timeout_count, char *e
           sprintf(strchr(answer, 0), " (%s)", p);
     }
 
-    strncpy(loganswer, answer, sizeof(loganswer) -1);
-    loganswer[sizeof(loganswer) -1] = 0;
-    cutspaces(loganswer);
-    cut_emptylines(loganswer);
+    snprintf(loganswer, sizeof(loganswer), "%s", answer);
 
-    // 3.1:
-    if (log_single_lines)
-      change_crlf(loganswer, ' ');
+    if (!log_unmodified)
+    {
+      cutspaces(loganswer);
+      cut_emptylines(loganswer);
+
+      if (log_single_lines)
+        change_crlf(loganswer, ' ');
+    }
 
     writelogfile(LOG_DEBUG, 0, "<- %s", loganswer);
   }
@@ -786,6 +934,9 @@ void setmodemparams()
   struct termios newtio;
   int baudrate;
 
+  if (DEVICE_IS_SOCKET)
+    return;
+
   bzero(&newtio, sizeof(newtio));
   newtio.c_cflag = CS8 | CLOCAL | CREAD | O_NDELAY | O_NONBLOCK;
   if (DEVICE.rtscts)
@@ -896,6 +1047,7 @@ int initmodem(char *new_smsc, int receiving)
   int retries=0;
   char *p;
   char *pre_initstring = "ATE0+CMEE=1\r";
+  char *pre_initstring_clip = "ATE0+CMEE=1;+CLIP=1\r";
   static int reading_checked = 0;
 
   STATISTICS->last_init = time(0);
@@ -904,6 +1056,17 @@ int initmodem(char *new_smsc, int receiving)
 
   // -----------------------------------------------------------------------------------------------
   writelogfile(LOG_INFO, 0, "Checking if modem is ready");
+
+  // 3.1.7: After being idle, some modems do not answer to the first AT command.
+  // With BenQ M32, there can be OK answer, but in many times there is not.
+  // To avoid error messages, first send AT and read the answer if it's available.
+  if (DEVICE.needs_wakeup_at)
+  {
+    put_command("AT\r", 0, 0, 1, 0);
+    usleep(100000);
+    read_from_modem(answer, sizeof(answer), 2);
+  }
+
   retries=0;
   do
   {
@@ -922,6 +1085,17 @@ int initmodem(char *new_smsc, int receiving)
       if (terminate)
         return 7;
     }
+
+    // 3.1.7: If it looks like modem does not respond, try to close and open the port:
+    if (retries >= 5 && !strstr(answer, "OK"))
+    {
+      try_closemodem(1);
+      sleep(1);
+
+      // If open fails, nothing can be done. Error is already logged. Will return 1.
+      if (!try_openmodem())
+        break;
+    }
   }
   while (retries <= 10 && !strstr(answer,"OK"));
   if (!strstr(answer,"OK"))
@@ -939,7 +1113,7 @@ int initmodem(char *new_smsc, int receiving)
   if (DEVICE.pre_init > 0)
   {
     writelogfile(LOG_INFO, 0, "Pre-initializing modem");
-    put_command(pre_initstring, answer, sizeof(answer), 2, "(OK)|(ERROR)");
+    put_command((DEVICE.phonecalls == 2)? pre_initstring_clip : pre_initstring, answer, sizeof(answer), 2, "(OK)|(ERROR)");
     if (!strstr(answer,"OK"))
       writelogfile(LOG_ERR, 1, "Modem did not accept the pre-init string");
   }
@@ -954,6 +1128,21 @@ int initmodem(char *new_smsc, int receiving)
     writelogfile(LOG_INFO, 0, "Checking if modem needs PIN");
     // 3.1.5: timeout from 50 to 100:
     put_command("AT+CPIN?\r", answer, sizeof(answer), 2, cpin_expect);
+
+    // 3.1.7: Some modems include quotation marks in the answer, like +CPIN: "READY".
+    while ((p = strchr(answer, '"')))
+      memmove(p, p +1, strlen(p +1) +1);
+
+    // 3.1.7: Some modems may leave a space away after +CPIN:
+    if (!strncmp(answer, "+CPIN:", 6) && strncmp(answer, "+CPIN: ", 7))
+    {
+      if ((p = strdup(answer)))
+      {
+        snprintf(answer, sizeof(answer), "+CPIN: %s", p + 6);
+        free(p);
+      }
+    }
+
     if (strstr(answer,"+CPIN: SIM PIN") && !strstr(answer, "PIN2"))
     {
       // 3.1.1:
@@ -1061,7 +1250,7 @@ int initmodem(char *new_smsc, int receiving)
     }
     // 3.1.5:
     else
-      explain_csq(LOG_NOTICE, 0, answer);
+      explain_csq(LOG_INFO, 0, answer);
   }
 
   // -----------------------------------------------------------------------------------------------
@@ -1087,6 +1276,9 @@ int initmodem(char *new_smsc, int receiving)
         STATISTICS->ber = atoi(p +1);
       else
         STATISTICS->ber = -2;
+
+      // 3.1.7: Explain signal quality to the log:
+      explain_csq(LOG_INFO, 0, answer);
     }
     else
     {
@@ -1137,7 +1329,14 @@ int initmodem(char *new_smsc, int receiving)
   if (new_smsc[0] || DEVICE.smsc[0])
   {
     writelogfile(LOG_INFO, 0, "Changing SMSC");
-    sprintf(command, "AT+CSCA=\"+%s\"\r", (new_smsc[0])? new_smsc : DEVICE.smsc);
+
+    // 3.1.7: clean + character(s) from the setting:
+    //sprintf(command, "AT+CSCA=\"+%s\"\r", (new_smsc[0]) ? new_smsc : DEVICE.smsc);
+    snprintf(answer, sizeof(answer), "%s", (new_smsc[0]) ? new_smsc : DEVICE.smsc);
+    while (*answer == '+')
+      memmove(answer, answer + 1, strlen(answer + 1) + 1);
+    sprintf(command, "AT+CSCA=\"+%s\"\r", answer);
+
     retries=0;
     do
     {
@@ -1170,7 +1369,7 @@ int initmodem(char *new_smsc, int receiving)
     // 3.1.5: do not remove ERROR text:
     if (!strstr(DEVICE.identity, "ERROR"))
       while (DEVICE.identity[0] && !isdigit(DEVICE.identity[0]))
-        strcpy(DEVICE.identity, DEVICE.identity +1);
+        memmove(DEVICE.identity, DEVICE.identity +1, strlen(DEVICE.identity +1) +1);
 
     // 3.1beta7: If CIMI is not supported, try CGSN (Product Serial Number)
     // TODO: is IMSI title still good?
@@ -1181,7 +1380,7 @@ int initmodem(char *new_smsc, int receiving)
 
       // 3.1:
       while (DEVICE.identity[0] && !isdigit(DEVICE.identity[0]))
-        strcpy(DEVICE.identity, DEVICE.identity +1);
+        memmove(DEVICE.identity, DEVICE.identity +1, strlen(DEVICE.identity +1) +1);
     }
     else
     {
@@ -1191,7 +1390,7 @@ int initmodem(char *new_smsc, int receiving)
 
       if (!strstr(answer, "ERROR"))
         while (answer[0] && !isdigit(answer[0]))
-          strcpy(answer, answer +1);
+          memmove(answer, answer +1, strlen(answer +1) +1);
 
       if ((p = strstr(answer, "OK")))
         *p = 0;
@@ -1214,7 +1413,8 @@ int initmodem(char *new_smsc, int receiving)
 
   // -----------------------------------------------------------------------------------------------
   // 3.1.5: Check once if reading of messages is not supported:
-  if (!reading_checked)
+  // 3.1.7: Do not check if not reading incoming messages:
+  if (DEVICE.incoming && !reading_checked)
   {
     reading_checked = 1;
     writelogfile(LOG_INFO, 0, "Checking if reading of messages is supported");
@@ -1235,6 +1435,54 @@ int initmodem(char *new_smsc, int receiving)
     }
   }
 
+  // -----------------------------------------------------------------------------------------------
+  // 3.1.7: Report details of device once:
+  if (DEVICE.report_device_details)
+  {
+    int save_log_single_lines = log_single_lines;
+    int i;
+    char tmp[256];
+    char *commands[] = {
+      "AT+CGMI", "Manufacturer identification",
+      "AT+CGMM", "Model identification",
+      "AT+CGMR", "Revision identification",
+      "AT+CNMI=?", "New message indications, list of supported modes",
+      "AT+CNMI?", "New message indications, current settings",
+      "AT+CPMS=?", "Preferred message storage, list of supported mem's",
+      //"AT+CPMS?", "Preferred message storage, current mem's and counters",
+      "AT+CPBS=?", "Phonebook storage, available mem's",
+      //"AT+CPBS?", "Phonebook storage, current storage and counters",
+      "AT+CMGL=?", "List messages, list of supported stat's",
+      "AT+CMGD=?", "Delete message, list of supported values",
+      "AT+CPAS=?", "Phone activity status, list of supported stat's",
+      "AT+CSCS=?", "TE character set, list of supported charset's",
+      "AT+CSCS?", "TE character set, current setting",
+      "" // end marker
+    };
+
+    DEVICE.report_device_details = 0;
+    log_single_lines = 0;
+    change_loglevel(LOG_DEBUG);
+
+    writelogfile(LOG_DEBUG, 0, "## Start of device details");
+
+    for (i = 0; commands[i][0]; i += 2)
+    {
+      if (terminate)
+        break;
+
+      snprintf(tmp, sizeof(tmp), "# %s:", commands[i + 1]);
+      writelogfile(LOG_DEBUG, 0, tmp);
+      sprintf(command, "%s\r", commands[i]);
+      put_command0(command, answer, sizeof(answer), 1, "(OK)|(ERROR)", 1);
+    }
+
+    writelogfile(LOG_DEBUG, 0, "## End of device details");
+
+    log_single_lines = save_log_single_lines;
+    restore_loglevel();
+  }
+
   // -----------------------------------------------------------------------------------------------
   // TODO: Check if AT+CMGD=? is supported.
 
@@ -1251,20 +1499,167 @@ int initialize_modem_receiving()
   return initmodem("", 1);
 }
 
-int openmodem()
-{
+#ifndef DISABLE_INET_SOCKET
 
-  modem_handle = open(DEVICE.device, O_RDWR | O_NOCTTY | O_NONBLOCK);
-  if (modem_handle < 0)
+/* Start Changes by Hubert Gilch, SEP Logistik AG
+ *
+ * 2 functions for connecting to a socket instead of a serial device
+ * in order to use ethernet GPRS-modems
+ *
+ * Code was "stolen" from interceptty by Scott W. Gifford
+ *
+ */
+
+struct sockaddr_in inet_resolve(const char *sockname)
+{
+  struct sockaddr_in sa;
+  char *hostname, *netport;
+  struct hostent *he;
+
+  if (!(hostname = strdup(sockname)))
+    writelogfile(LOG_CRIT, 0, "Couldn't dup string: %s", strerror(errno));
+
+  netport = strchr(hostname, ':');
+  *netport = '\0';
+  netport++;
+
+  sa.sin_family = AF_INET;
+
+  if (!(he = gethostbyname(hostname)))
+    writelogfile(LOG_ERR, 0, "Couldn't resolve name '%s': %s.", hostname,
+      (h_errno == HOST_NOT_FOUND) ? "Host not found" :
+      ((h_errno == NO_ADDRESS) || (h_errno == NO_DATA)) ? "No data available" :
+      (h_errno == NO_RECOVERY) ? "A non-recoverable name server error occured" : (h_errno == TRY_AGAIN) ? "A temporary error occured." : "An unknown error occured");
+
+  memcpy(&(sa.sin_addr), he->h_addr, he->h_length);
+
+#if 0
+  if (!(se = getservbyname(netport)))
+    writelogfile(LOG_ERR, 0, "Couldn't resolve port.");
+
+  host_port = htons(se->s_port);
+#endif
+
+  if (!(sa.sin_port = htons(atoi(netport))))
+    writelogfile(LOG_ERR, 0, "Couldn't figure out port number.");
+
+  free(hostname);
+
+  return sa;
+}
+
+int open_inet_socket(char *backend)
+{
+  struct sockaddr_in sa;
+  int fd;
+  int socketflags;
+  int retries = 0;
+
+  sa = inet_resolve(backend + 1);	// cut first character @
+  if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 3)
   {
-    writelogfile0(LOG_ERR, 1, tb_sprintf("Cannot open serial port %s, error: %s",
-                  DEVICE.device, strerror(errno)));
+    tb_sprintf("Couldn't open socket: %s: %s", backend, strerror(errno));
+    writelogfile0(LOG_ERR, 0, tb);
     alarm_handler0(LOG_ERR, tb);
     return -1;
   }
+
+  while (connect(fd, (struct sockaddr *) &sa, sizeof(sa)) != 0)
+  {
+    retries++;
+
+    if (terminate || (DEVICE.socket_connection_retries != -1 && retries > DEVICE.socket_connection_retries))
+    {
+      close(fd);
+      return (terminate)? -2 : -1;
+    }
+
+    tb_sprintf("Couldn't connect socket %s, error: %s, waiting %i sec.", backend, strerror(errno), DEVICE.socket_connection_errorsleeptime);
+
+    if (retries - 1 == DEVICE.socket_connection_alarm_after)
+    {
+      writelogfile0(LOG_ERR, 0, tb);
+      alarm_handler0(LOG_ERR, tb);
+    }
+    else
+    {
+      // Do not log the first failure:
+      if (retries > 1)
+        writelogfile(LOG_INFO, 0, tb);
+    }
+
+    sleep(DEVICE.socket_connection_errorsleeptime);
+  }
+
+  socketflags = fcntl(fd, F_GETFL);
+  fcntl(fd, F_SETFL, socketflags | O_NONBLOCK);
+
+  return fd;
+}
+
+/*
+ * End Changes by Hubert Gilch
+ */
+
+#endif
+
+int openmodem()
+{
+  int retries = 0;
+
+  // 3.1.7:
+  //modem_handle = open(DEVICE.device, O_RDWR | O_NOCTTY | O_NONBLOCK);
+
+  /*
+   * if devicename starts with "@" it is not a serial device but
+   * a socket, so open a socket instead a device file
+   *
+   * Change by Hubert Gilch, SEP Logistik AG
+   */
+#ifndef DISABLE_INET_SOCKET
+  if (DEVICE_IS_SOCKET)
+    modem_handle = open_inet_socket(DEVICE.device);
+  else
+#endif
+    // 3.1.7:
+    while ((modem_handle = open(DEVICE.device, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0)
+    {
+      retries++;
+
+      if (terminate || (DEVICE.device_open_retries != -1 && retries > DEVICE.device_open_retries))
+        break;
+
+      tb_sprintf("Couldn't open serial port %s, error: %s, waiting %i sec.", DEVICE.device, strerror(errno), DEVICE.device_open_errorsleeptime);
+
+      if (retries - 1 == DEVICE.device_open_alarm_after)
+      {
+        writelogfile0(LOG_ERR, 0, tb);
+        alarm_handler0(LOG_ERR, tb);
+      }
+      else
+        writelogfile(LOG_INFO, 0, tb);
+
+      sleep(DEVICE.device_open_errorsleeptime);
+    }
+
+  if (modem_handle < 0)
+  {
+    if (modem_handle == -1)
+    {
+      writelogfile0(LOG_ERR, 1, tb_sprintf((DEVICE_IS_SOCKET)? "Cannot open socket %s, error: %s" : "Cannot open serial port %s, error: %s", DEVICE.device, strerror(errno)));
+      alarm_handler0(LOG_ERR, tb);
+    }
+    return -1;
+  }
+
   if (strstr(smsd_version, "beta"))
-    writelogfile(LOG_INFO, 0, "Serial port %s opened as %i, rtscts: %i, baudrate: %i",
-                 DEVICE.device, modem_handle, DEVICE.rtscts, DEVICE.baudrate);
+  {
+    if (DEVICE_IS_SOCKET)
+      writelogfile(LOG_INFO, 0, "Socket %s opened as %i", DEVICE.device, modem_handle);
+    else
+      writelogfile(LOG_INFO, 0, "Serial port %s opened as %i, rtscts: %i, baudrate: %i", DEVICE.device, modem_handle, DEVICE.rtscts, DEVICE.baudrate);
+  }
+
   return modem_handle;
 }
 
@@ -1418,7 +1813,7 @@ int wait_network_registration(int waitnetwork_errorsleeptime, int retry_count)
 
     // 3.1.1: Some modem include spaces in the response:
     while ((p = strchr(answer, ' ')))
-      strcpy(p, p +1);
+      memmove(p, p +1, strlen(p +1) +1);
 
     // 3.1.1: Drop additional fields:
     if ((p = strchr(answer, ',')))
@@ -1427,7 +1822,7 @@ int wait_network_registration(int waitnetwork_errorsleeptime, int retry_count)
 
     // 3.1.1: Some modem (Motorola) gives values using three digits like "000,001":
     if ((p = strstr(answer, ",00")))
-      strcpy(p +1, p +3);
+      memmove(p +1, p +3, strlen(p +3) +1);
 
     // 3.1:
     // Second field is tested.
@@ -1509,13 +1904,13 @@ int try_closemodem(int force)
 
   if (modem_handle >= 0 && !keep_open)
   {
-    if (0 && strstr(smsd_version, "beta"))
+    if (1 && strstr(smsd_version, "beta"))
     {
-      writelogfile(LOG_INFO, 0, "Serial port %s (%i) closed", DEVICE.device, modem_handle);
+      writelogfile(LOG_INFO, 0, "Device %s (%i) closed", DEVICE.device, modem_handle);
       writelogfile(LOG_INFO, 0, "***********");
     }
 #ifdef DEBUGMSG
-  printf("!! Closing serial port %s\n", DEVICE.device);
+  printf("!! Closing device %s\n", DEVICE.device);
 #endif 
     close(modem_handle);
     modem_handle = -2;
@@ -1531,13 +1926,13 @@ int try_openmodem()
   if (modem_handle >= 0)
   {
 #ifdef DEBUGMSG
-  printf("!! Opening serial port %s: already open\n", DEVICE.device);
+  printf("!! Opening device %s: already open\n", DEVICE.device);
 #endif 
     return 1;
   }
 
 #ifdef DEBUGMSG
-  printf("!! Opening serial port %s\n", DEVICE.device);
+  printf("!! Opening device %s\n", DEVICE.device);
 #endif 
   if (openmodem() == -1)
   {
