@@ -169,7 +169,7 @@ int text2pdu(char* text, int length, char* pdu, char* udh)
   if (length>maxsms_pdu-udh_size_septets)
     length=maxsms_pdu-udh_size_septets;
   //clear the tmp buffer
-  for (character=0;character<sizeof(tmp);character++)
+  for (character=0;(size_t)character<sizeof(tmp);character++)
     tmp[character]=0;
   // Convert 8bit text stream to 7bit stream
   for (character=0;character<length;character++)
@@ -350,7 +350,7 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
     {
       strcpy(tmp2, udh_data);
       while ((p = strchr(tmp2, ' ')))
-        strcpy(p, p +1);
+        memmove(p, p +1, strlen(p +1) +1);
       l = strlen(tmp2) /2;
       binary2pdu(message, messagelen, strchr(tmp2, 0));
       messagelen += l;
@@ -360,18 +360,38 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
   }
   else
     messagelen=text2pdu(message,messagelen,tmp2,udh_data);
+
   /* concatenate the first part of the PDU string */
   if (strcmp(mode,"old")==0)
     sprintf(pdu,"%02X00%02X%02X%s00%02X%02X",flags,numberlength,numberformat,tmp,coding,messagelen);
   else
   {
+    int proto = 0;
+
     if (validity < 0 || validity > 255)
       validity = validity_period;
-    sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X",
-            flags, numberlength, numberformat, tmp,
-            (system_msg)? 0x40 : (replace_msg >= 1 && replace_msg <= 7)? 0x40 +replace_msg : 0,
-            (system_msg)? 0xF4 : coding, validity, messagelen);
+
+    if (system_msg)
+    {
+      proto = 0x40;
+      coding = 0xF4;  // binary
+
+      // 3.1.7:
+      if (system_msg == 2)
+      {
+        proto += (0x7F - 0x40); // SS (no show)
+        coding += 2;    // store to sim
+      }
+    }
+    else if (replace_msg >= 1 && replace_msg <= 7)
+       proto = 0x40 + replace_msg;
+
+    //sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp,
+    // (system_msg) ? 0x40 : (replace_msg >= 1 && replace_msg <= 7) ? 0x40 + replace_msg : 0,
+    // (system_msg) ? 0xF4 : coding, validity, messagelen);
+    sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp, proto, coding, validity, messagelen);
   }
+
   /* concatenate the text to the PDU string */
   strcat(pdu,tmp2);
 }
@@ -530,12 +550,12 @@ int explain_udh(char *udh_type, char *pdu)
     return -1;
   strcpy(buffer, pdu);
   while ((p = strchr(buffer, ' ')))
-    strcpy(p, p +1);
+    memmove(p, p +1, strlen(p +1) +1);
 
   if ((udh_length = octet2bin_check(buffer)) < 0)
     return -1;
   udh_length++;
-  if (udh_length *2 > strlen(buffer))
+  if ((size_t)(udh_length *2) > strlen(buffer))
     return -1;
   sprintf(udh_type, "Length=%i", udh_length);
   idx = 1;
@@ -612,7 +632,7 @@ int explain_udh(char *udh_type, char *pdu)
     // Next octet is length of data:
     if ((i = octet2bin_check(Src_Pointer +2)) < 0)
       return -1;
-    if (i *2 > strlen(Src_Pointer +4))
+    if ((size_t)(i *2) > strlen(Src_Pointer +4))
       return -1;
     idx += i +2;
     if (idx > udh_length)
@@ -641,7 +661,7 @@ int pdu2text(char *pdu, char *text, int *text_length, int *expected_length,
   int octetcounter;
   int skip_characters = 0;
   char c;
-  char binary;
+  char binary = 0;
   int i;
   int result;
 
@@ -973,7 +993,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
         Src_Pointer += 2;
         if ((i & 112) == 80)
         {  // Sender is alphanumeric
-          if (strlen(Src_Pointer) < Length +padding)
+          if (strlen(Src_Pointer) < (size_t)(Length +padding))
             pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                       "While trying to read sender address (alphanumeric, length %i): %s",
                       Length +padding, err_too_short);
@@ -987,7 +1007,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
         }
         else
         {  // sender is numeric
-          if (strlen(Src_Pointer) < Length +padding)
+          if (strlen(Src_Pointer) < (size_t)(Length +padding))
             pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                       "While trying to read sender address (numeric, length %i): %s",
                       Length +padding, err_too_short);
@@ -1196,7 +1216,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
           Src_Pointer += 2;
           if ((i & 112) == 80)
           {  // Sender is alphanumeric
-            if (strlen(Src_Pointer) < Length +padding)
+            if (strlen(Src_Pointer) < (size_t)(Length +padding))
               pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                         "While trying to read recipient address (alphanumeric, length %i): %s",
                         Length +padding, err_too_short);
@@ -1210,7 +1230,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
           }
           else
           {  // sender is numeric
-            if (strlen(Src_Pointer) < Length +padding)
+            if (strlen(Src_Pointer) < (size_t)(Length +padding))
               pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                         "While trying to read recipient address (numeric, length %i): %s",
                         Length +padding, err_too_short);
@@ -1427,7 +1447,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
   // Patch for Wavecom SR memory reading:
   if (strncmp(pdu, "000000FF00", 10) == 0)
   {
-    strcpy(pdu, pdu +8);
+    memmove(pdu, pdu +8, strlen(pdu +8) +1);
     while (strlen(pdu) < 52)
       strcat(pdu, "00");
   }
@@ -1492,7 +1512,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
             {
               Length = Length *2 -2;
               // No padding because the given value is number of octets.
-              if (strlen(Pointer) < Length +4)
+              if (strlen(Pointer) < (size_t)(Length +4))
                 pdu_error(&err_str, 0, Pointer -pdu, Length +4, "While trying to read sender SMSC address (length %i): %s",
                           Length, err_too_short);
               else
@@ -1590,8 +1610,12 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
                 else
                 {
                   memcpy(ascii, message, message_length);
+#ifndef USE_ICONV
                   ascii[message_length] = 0;
                   i = decode_ucs2(ascii, message_length);
+#else
+                  i = (int)iconv_ucs2utf(ascii, message_length, sizeof(ascii));
+#endif
                   expected_length /= 2;
                 }
 
@@ -1674,7 +1698,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
   return result;
 }
 
-int get_pdu_details(char *dest, int size_dest, char *pdu, int mnumber)
+int get_pdu_details(char *dest, size_t size_dest, char *pdu, int mnumber)
 {
   int result = 0;
   int udlen;
@@ -1696,7 +1720,7 @@ int get_pdu_details(char *dest, int size_dest, char *pdu, int mnumber)
   int flash;
   int bin_udh = 1;
   int m_id = 999; // real id for concatenated messages only, others use this.
-  int length_sender = 32;
+  size_t length_sender = 32;
   int p_count = 1;
   int p_number = 1;
   char buffer[100]; // 3.1.6: increased size, was too small (51).
@@ -1759,9 +1783,32 @@ int get_pdu_details(char *dest, int size_dest, char *pdu, int mnumber)
       }
 /*
 001 A 358401234567____________________111 001/001 r 00-00-00 00-00-00n
+mnumber
+    sort_ch
+      sender
+                                      m_id
+                                          p_number
+                                              p_count
+                                                  incoming / report
+                                                    date
+                                                             time
+
+      snprintf(buffer, sizeof(buffer), "%.03i %c %s%.03i %.03i/%.03i %c %-8.8s %-8.8s\n",
+               mnumber, sort_ch, sender, m_id, p_number, p_count, (is_statusreport) ? 'r' : 'i', date, time);
+
+3.1.7:
+001 A 00-00-00 00-00-00 358401234567____________________111 001/001 rn
+mnumber
+    sort_ch
+      date
+               time
+                        sender
+                                                        m_id
+                                                            p_number
+                                                                p_count
+                                                                    incoming / report
 */
-      sprintf(buffer, "%.03i %c %s%.03i %.03i/%.03i %c %-8.8s %-8.8s\n",
-              mnumber, sort_ch, sender, m_id, p_number, p_count, (is_statusreport)? 'r' : 'i', date, time);
+      snprintf(buffer, sizeof(buffer), "%.03i %c %-8.8s %-8.8s %s%.03i %.03i/%.03i %c\n", mnumber, sort_ch, date, time, sender, m_id, p_number, p_count, (is_statusreport) ? 'r' : 'i');
 
       if (strlen(dest) +strlen(buffer) < size_dest)
         strcat(dest, buffer);
@@ -1778,7 +1825,8 @@ int get_pdu_details(char *dest, int size_dest, char *pdu, int mnumber)
 
 int sort_pdu_helper(const void *a, const void *b)
 {
-  return(strncmp((char *)a +4, (char *)b +4, 45));
+  //return(strncmp((char *)a +4, (char *)b +4, 45));
+  return (strncmp((char *) a + 4, (char *) b + 4, 63));
 }
 
 void sort_pdu_details(char *dest)
