@@ -20,16 +20,28 @@ Either version 2 of the License, or (at your option) any later version.
 #include <sys/types.h>
 #include <time.h>
 
+#ifndef __FreeBSD__
+#define DEFAULT_CONFIGFILE "/etc/smsd.conf"
+#else
+#define DEFAULT_CONFIGFILE "%%PREFIX%%/etc/smsd.conf"
+#endif
+
 #define DATETIME_DEFAULT "%y-%m-%d %H:%M:%S"
+#define LOGTIME_DEFAULT "%Y-%m-%d %H:%M:%S"
+#define DATE_FILENAME_DEFAULT "%Y-%m-%d"
 
 #define CONCATENATED_DIR_FNAME "%s/%s-concatenated"
 
 #define MM_CORE_FNAME "/tmp/mm_smsd_%i" /* %i is PID */
 
 #define NUMS 64
+#define SIZE_NUM 16
 
 #define DEVICE devices[process_id]
+#define DEVICE_IS_SOCKET (devices[process_id].device[0] == '@')
+#define DEVICE_X_IS_SOCKET (devices[x].device[0] == '@')
 #define STATISTICS statistics[process_id]
+#define strcpyo(dest,src) memmove(dest, src, strlen(src) +1)
 
 // Maximum size of a message text
 #define MAXTEXT 39016
@@ -57,6 +69,8 @@ Either version 2 of the License, or (at your option) any later version.
 #define SIZE_PRIVILEDGED_NUMBERS 512
 #define SIZE_SMSD_DEBUG 100
 #define SIZE_SHARED_BUFFER 256
+#define SIZE_FILENAME_PREVIEW 256
+#define SIZE_PB_ENTRY 101
 
 // Check memory methods:
 #define CM_NO_CPMS 0
@@ -91,7 +105,7 @@ int put_command_timeouts;
 typedef struct
 {
   char name[32]; 		// Name of the queue
-  char numbers[NUMS][16];	// Phone numbers assigned to this queue
+  char numbers[NUMS][SIZE_NUM];	// Phone numbers assigned to this queue
   char directory[PATH_MAX];		// Queue directory
 } _queue;
 
@@ -100,6 +114,9 @@ typedef struct
   char name[32];		// Name of the modem
   char number[32];              // 3.1.4: SIM card's telephone number.
   char device[PATH_MAX];	// Serial port name
+  int device_open_retries;      // 3.1.7: Defines count of retries when opening a device fails.
+  int device_open_errorsleeptime; // 3.1.7: Sleeping time after opening error.
+  int device_open_alarm_after;  // 3.1.7: Defines after how many failures an alarmhandler is called.
   char identity[SIZE_IDENTITY]; // Identification asked from the modem (CIMI)
   char conf_identity[SIZE_IDENTITY]; // Identification set in the conf file (CIMI)
   //char identity_header[SIZE_TO];// Title of current identification (IMSI)
@@ -107,17 +124,22 @@ typedef struct
   int incoming; 		// Try to receive SMS. 0=No, 1=Low priority, 2=High priority
   int outgoing;                 // = 0 if a modem is not used to send messages.
   int report; 			// Ask for delivery report 0 or 1 (experimental state)
-  int phonecalls;               // Check for phonebook status for calls, 0 or 1
+  int phonecalls;               // Check for phonebook status for calls, 0 or 1. 3.1.7: value 2 = +CLIP report is handled.
+  char phonecalls_purge[32];    // Defines if a purge command should be used. yes / no / command to use. yes = AT^SPBD="MC"
+  int phonecalls_error_max;     // 3.1.7: Max nr of errors before phonecalls are ignored.
   char pin[16];			// SIM PIN
   int pinsleeptime;             // Number of seconds to sleep after a PIN is entered.
   char mode[10]; 		// Command set version old or new
   char smsc[16];		// Number of SMSC
   int baudrate;			// Baudrate
   int send_delay;		// Makes sending characters slower (milliseconds)
+  int send_handshake_select;    // 3.1.9.
   int cs_convert; 		// Convert character set  0 or 1 (iso-9660)
   char initstring[100];		// first Init String
   char initstring2[100];        // second Init String
   char eventhandler[PATH_MAX];	// event handler program or script
+  char eventhandler_ussd[PATH_MAX]; // 3.1.7: event handler program or script for USSD answers
+  int ussd_convert;             // 3.1.7: Convert string from USSD answer
   int rtscts;			// hardware handshake RTS/CTS, 0 or 1
   int read_memory_start;	// first memory space for sms
   char primary_memory[10];      // primary memory, if dual-memory handler is used
@@ -138,6 +160,7 @@ typedef struct
   int message_count_clear;      // Period to automatically clear message counter. Value is MINUTES. 
   int keep_open;                // 1 = do not close modem while idle.
   char dev_rr[PATH_MAX];        // Script/program which is run regularly.
+  char dev_rr_post_run[PATH_MAX]; // 3.1.7: Script/program which is run regularly (POST_RUN).
   int dev_rr_interval;          // Number of seconds between running a regular_run script/progdam.
   char dev_rr_cmdfile[PATH_MAX];// 
   char dev_rr_cmd[SIZE_RR_CMD]; //
@@ -150,6 +173,8 @@ typedef struct
   int voicecall_vts_list;       // Defines how VTS command is sent: 1 = as a list like "1,2,3,4", 2 = single note with one VTS command (default).
   int voicecall_ignore_modem_response; // Delay defined with TIME: is not breaked even if modem gives some response.
   int voicecall_hangup_ath;     // If ATH is used instead of AT+CHUP.
+  int voicecall_vts_quotation_marks; // Defines if AT+VTS="n" command is given with quotation marks.
+  int voicecall_cpas;           // Defines if AT+CPAS is used to detect when a call is answered (phone returns OK after ATD).
   int check_memory_method;      // 0 = CPMS not supported, 1 = CPMS supported and must work (default), 2 = CMGD used to check messages, 3 = CMGL is used.
   char cmgl_value[32];          // With check_memory_method 3, correct value for AT+CMGL= must be given here.
   char priviledged_numbers[SIZE_PRIVILEDGED_NUMBERS]; // Priviledged numbers in incoming messages.
@@ -167,6 +192,18 @@ typedef struct
   int communication_delay;      // Time between each put_command (milliseconds), some modems need this.
   int hangup_incoming_call;     // 1 = if detected unexpected input contains RING and we want to end call.
   int max_continuous_sending;   // Defines when sending is breaked to do check/do other tasks. Time in seconds.
+  int socket_connection_retries; // 3.1.7: Defines count of retries when socket connection fails.
+  int socket_connection_errorsleeptime; // 3.1.7: Sleeping time after socket connetcion error.
+  int socket_connection_alarm_after; // 3.1.7: Defines after how many failures an alarmhandler is called.
+  int report_device_details;    // Defines if device details are logged when modem process is starting.
+  int using_routed_status_report; // Disables a warning about routed status reports.
+  int routed_status_report_cnma; // Defines if +CNMA acknowledgement is needed to send.
+  int needs_wakeup_at;          // After idle time, some modems may not answer to the first AT command.
+  int keep_messages;            // Defines if messages are not deleted. Smsd continues running.
+  char startstring[100];        // 3.1.7: Command(s) to send to the modem when a devicespooler is starting.
+  int startsleeptime;           // 3.1.7: Second to wait after startstring is sent.
+  char stopstring[100];         // 3.1.7: Command(s) to send to the modem when a devicespooler is stopping.
+  int trust_spool;		// 3.1.9
 } _device;
 
 // NOTE for regular run intervals: effective value is at least delaytime.
@@ -235,6 +272,31 @@ int trust_outgoing;             // 1 = it's _sure_ that files are created by ren
 // Undocumented:
 int ignore_outgoing_priority;   // 1 = Priority: high header is not checked. Speeds up spooling.
 
+// 3.1.7:
+int ignore_exec_output;         // 1 = stdout and stderr of eventhandlers is _not_ checked.
+
+// 3.1.7:
+mode_t conf_umask;              // File mode creation mask for smsd and modem processes.
+
+// 3.1.7:
+int trim_text;                  // 1 = trailing whitespaces are removed from text:
+
+// 3.1.7:
+int use_linux_ps_trick;         // 1 = change argv[0] to "smsd: MAINPROCESS", "smsd: GSM1" etc.
+
+// 3.1.7:
+int log_unmodified;
+
+// 3.1.7:
+char suspend_filename[PATH_MAX];
+
+// Undocumented:
+// 3.1.9:
+int spool_directory_order;
+
+// 3.1.9: 1 if read_from_modem is logged.
+int log_read_from_modem;
+
 int message_count;              // Counter for sent messages. Multipart message is one message.
 
 int terminate;                  // The current process terminates if this is 1
@@ -252,6 +314,9 @@ char arg_infofile[PATH_MAX];
 char arg_pidfile[PATH_MAX];
 char arg_logfile[PATH_MAX];
 int arg_terminal;
+// 3.1.7:
+char arg_7bit_packed[512];
+int do_encode_decode_arg_7bit_packed;
 
 int terminal;                   // 1 if smsd is communicating with terminal.
 pid_t device_pids[NUMBER_OF_MODEMS]; // Pid's of modem processes.
@@ -288,6 +353,8 @@ char no_chars[SIZE_HEADER];     // See details inside read_translation() functio
 char yes_word[SIZE_HEADER];     // "yes" printed as an output.
 char no_word[SIZE_HEADER];      // "no"
 char datetime_format[SIZE_HEADER]; // strftime format string for time stamps (not inside status reports).
+char logtime_format[SIZE_HEADER]; // 3.1.7: strftime format string for logging time stamps
+char date_filename_format[SIZE_HEADER]; // 3.1.7: strftime format string for date_filename
 int translate_incoming;         // 0 if incoming message headers are NOT transtaled.
 
 // Next two are ffor debugging purposes:
@@ -329,4 +396,6 @@ void abnormal_termination(int all);
 
 char *tb_sprintf(char* format, ...);
 
+int savephonecall(char *entry_number, int entry_type, char *entry_text);
+
 #endif
