@@ -263,7 +263,7 @@ void checkwritestats()
       last_stats=time(0);
       // %Y used instead of %y to avoid compiler warning message in some environments.
       strftime(s,sizeof(s),"%Y%m%d.%H%M%S",localtime(&next_time));
-      strcpy(s, s +2);
+      memmove(s, s +2, strlen(s +2) +1);
       syslog(LOG_INFO,"Writing stats file %s",s);
       strcpy(filename,d_stats);
       strcat(filename,"/");
@@ -328,15 +328,14 @@ void update_message_counter(int messages, char *modemname)
       fclose(fp);
     }
 
-    if (messages > 0)
-    {
-      counter += messages;
+    // 3.1.7: always create a file
 
-      if ((fp = fopen(filename, "w")))
-      {
-        fprintf(fp, "%s: %i\n", modemname, counter);
-        fclose(fp);
-      }
+    counter += messages;
+
+    if ((fp = fopen(filename, "w")))
+    {
+      fprintf(fp, "%s: %i\n", modemname, counter);
+      fclose(fp);
     }
 
     STATISTICS->message_counter = counter;
@@ -354,7 +353,7 @@ void write_status()
   char *status;
   char buffer[256];
   time_t t;
-  static int longest_modemname = 0;
+  static size_t longest_modemname = 0;
   int include_counters;
 
   if (!printstatus && d_stats[0])
